from datasette import hookimpl
from datasette.utils.asgi import Response, Forbidden
import asyncio
import json
from pathlib import Path
import urllib


async def run_ripgrep(
    pattern,
    path,
    globs=None,
    time_limit=1.0,
    max_lines=2000,
    ignore=False,
    literal=False,
    context=2,
):
    args = ["-e", pattern, path, "--json"]
    if context:
        args.extend(["-C", str(context)])
    if ignore:
        args.append("-i")
    if literal:
        args.append("-F")
    if globs:
        for glob in globs:
            args.extend(["--glob", glob])
    proc = await asyncio.create_subprocess_exec(
        "rg",
        *args,
        stdout=asyncio.subprocess.PIPE,
        stdin=asyncio.subprocess.PIPE,
        limit=1024 * 1024,
        cwd=path,
    )
    max_lines_hit = False
    time_limit_hit = False
    results = []

    async def inner(results):
        nonlocal max_lines_hit
        while True:
            try:
                line = await proc.stdout.readline()
            except (asyncio.exceptions.LimitOverrunError, ValueError):
                # Skip 'Separator is not found, and chunk exceed the limit' lines
                continue
            if line == b"":
                break
            try:
                results.append(json.loads(line))
            except json.decoder.JSONDecodeError:
                # Usually this means a really long line which was
                # truncated as invalid JSON
                print(line)
            if len(results) >= max_lines:
                max_lines_hit = True
                break

    results = []

    try:
        await asyncio.wait_for(inner(results), timeout=time_limit)
    except asyncio.TimeoutError:
        time_limit_hit = True
    try:
        proc.kill()
    except OSError:
        # Ignore 'no such process' error
        pass
    # We should have accumulated some results anyway
    return results, time_limit_hit


async def ripgrep(request, datasette):
    await check_permission(request, datasette)
    pattern = (request.args.get("pattern") or "").strip()
    ignore = request.args.get("ignore")
    literal = request.args.get("literal")
    globs = [g.strip() for g in request.args.getlist("glob") if g.strip()]

    config = datasette.plugin_config("datasette-ripgrep") or {}
    time_limit = config.get("time_limit") or 1.0
    max_lines = config.get("max_lines") or 2000
    path = config.get("path")
    if not path:
        return Response.html("The path plugin configuration is required.", status=500)

    results = []
    time_limit_hit = False
    if pattern:
        results, time_limit_hit = await run_ripgrep(
            pattern,
            path,
            globs=globs,
            time_limit=time_limit,
            max_lines=max_lines,
            ignore=ignore,
            literal=literal,
        )

    def fix_path(path_):
        return str(Path(path_).relative_to(path))

    try:
        widest_line_number = len(
            str(
                max(
                    result["data"]["line_number"]
                    for result in results
                    if "line_number" in result["data"]
                )
            )
        )
    except ValueError:
        # max() arg is an empty sequence
        widest_line_number = 1

    return Response.html(
        await datasette.render_template(
            "ripgrep.html",
            {
                "pattern": pattern,
                "results": results,
                "fix_path": fix_path,
                "time_limit_hit": time_limit_hit,
                "url_quote": urllib.parse.quote,
                "literal": literal,
                "ignore": ignore,
                "globs": globs,
                "widest_line_number": widest_line_number,
            },
            request=request,
        )
    )


async def view_file(request, datasette):
    await check_permission(request, datasette)
    config = datasette.plugin_config("datasette-ripgrep") or {}
    subpath = urllib.parse.unquote(request.url_vars["subpath"])
    path = config.get("path")
    if not path:
        return Response.html("The path plugin configuration is required.", status=500)
    filepath = Path(path) / subpath
    filepath = filepath.resolve()
    # Make absolutely sure it's still inside the root
    if not str(filepath).startswith(str(path)):
        return Response.html("File must be inside path directory", status=403)
    if not filepath.exists():
        return Response.text("File not found: {}".format(subpath), status=404)
    lines = filepath.read_text().split("\n")
    widest_line_number = len(str(len(lines) + 1))
    return Response.html(
        await datasette.render_template(
            "ripgrep_view_file.html",
            {
                "subpath": subpath,
                "lines": enumerate(lines),
                "widest_line_number": widest_line_number,
            },
            request=request,
        )
    )


async def check_permission(request, datasette):
    can_view = await datasette.allowed(
        action="view-instance",
        actor=request.actor,
    )
    if not can_view:
        raise Forbidden("view-instance denied")


@hookimpl
def register_routes():
    return (
        ("^/-/ripgrep$", ripgrep),
        ("^/-/ripgrep/view/(?P<subpath>.*)$", view_file),
    )


@hookimpl
def menu_links(datasette, actor):
    config = datasette.plugin_config("datasette-ripgrep") or {}
    if not config.get("path"):
        return
    return [
        {"href": datasette.urls.path("/-/ripgrep"), "label": "ripgrep search"},
    ]



from datasette.app import Datasette
import pytest
import re
import shutil
import textwrap


@pytest.fixture(scope="session")
def src(tmp_path_factory):
    root = tmp_path_factory.mktemp("root")
    src = root / "src"
    src.mkdir()
    (src / "one.txt").write_text("Hello\nThere\nThis\nIs a.test file")
    (src / "sub").mkdir()
    (src / "sub/two.txt").write_text("Second test file")
    # This one is for testing class="gap"
    (src / "sub/three.txt").write_text("XXX\nYYY\nZZZ\nZZZ\nZZZ\nZZZ\nZZZ\nYYY")
    (src / "{{curlies}}.txt").write_text("File with curlies in the name -v")
    return src


@pytest.fixture(scope="session")
def datasette(src):
    return Datasette(
        [],
        memory=True,
        config={
            "plugins": {
                "datasette-ripgrep": {
                    "path": str(src),
                }
            }
        },
    )


@pytest.mark.asyncio
async def test_plugin_is_installed(datasette):
    response = await datasette.client.get("/-/plugins.json")
    assert 200 == response.status_code
    installed_plugins = {p["name"] for p in response.json()}
    assert "datasette-ripgrep" in installed_plugins


@pytest.mark.asyncio
@pytest.mark.skipif(not shutil.which("rg"), reason="rg executable not found")
@pytest.mark.parametrize(
    "querystring,expected_title,expected_fragments,unexpected_fragments",
    (
        (
            "pattern=est",
            "est",
            [
                # This one also tests the context
                (
                    "        <h3>one.txt</h3>\n"
                    '        <div style="overflow-x: auto">\n'
                    '        <pre class=""><a class="line-number" href="/-/ripgrep/view/one.txt#L2">2   </a><span>There</span></pre>\n'
                    '        <pre class=""><a class="line-number" href="/-/ripgrep/view/one.txt#L3">3   </a><span>This</span></pre>\n'
                    '        <pre class="match"><a class="line-number" href="/-/ripgrep/view/one.txt#L4">4   </a><span>Is a.test file</span></pre>\n'
                    "        </div>"
                ),
                (
                    "        <h3>sub/two.txt</h3>\n"
                    '        <div style="overflow-x: auto">\n'
                    '        <pre class="match"><a class="line-number" href="/-/ripgrep/view/sub/two.txt#L1">1   </a><span>Second test file</span></pre>\n'
                    "        </div>\n"
                ),
            ],
            [],
        ),
        ("pattern=EST", "EST", [], ["<h3>one.txt</h3>\n"]),
        (
            "pattern=EST&ignore=on",
            "EST",
            [
                "<h3>one.txt</h3>",
                '<pre class="match"><a class="line-number" href="/-/ripgrep/view/one.txt#L4">4   </a><span>Is a.test file</span></pre>',
            ],
            [],
        ),
        (
            "pattern=.test",
            ".test",
            [
                "<h3>one.txt</h3>",
                '<pre class="match"><a class="line-number" href="/-/ripgrep/view/one.txt#L4">4   </a><span>Is a.test file</span></pre>',
                # " test" matches regex ".test"
                "<h3>sub/two.txt</h3>",
                '<pre class="match"><a class="line-number" href="/-/ripgrep/view/sub/two.txt#L1">1   </a><span>Second test file</span></pre>',
            ],
            [],
        ),
        (
            "pattern=.test&literal=on",
            ".test",
            [
                "<h3>one.txt</h3>",
                '<pre class="match"><a class="line-number" href="/-/ripgrep/view/one.txt#L4">4   </a><span>Is a.test file</span></pre>',
            ],
            # " test" does not match literal ".test"
            [
                "<h3>sub/two.txt</h3>",
                '<pre class="match"><a class="line-number" href="/-/ripgrep/view/sub/two.txt#L1">1   </a><span>Second test file</span></pre>',
            ],
        ),
        (
            "pattern=test&glob=two.txt",
            "test",
            [
                "<h3>sub/two.txt</h3>",
                '<pre class="match"><a class="line-number" href="/-/ripgrep/view/sub/two.txt#L1">1   </a><span>Second test file</span></pre>',
            ],
            ["<h3>one.txt</h3>"],
        ),
        (
            "pattern=test&glob=!two.txt",
            "test",
            [
                "<h3>one.txt</h3>",
            ],
            ["<h3>sub/two.txt</h3>"],
        ),
        (
            "pattern=YYY",
            "YYY",
            [
                "<h3>sub/three.txt</h3>",
                '<pre class=" gap"><a class="line-number" href="/-/ripgrep/view/sub/three.txt#L6">',
            ],
            [],
        ),
    ),
)
async def test_ripgrep_search(
    datasette, querystring, expected_title, expected_fragments, unexpected_fragments
):
    response = await datasette.client.get("/-/ripgrep?{}".format(querystring))
    assert "<title>ripgrep: {}</title>".format(expected_title) in response.text
    html = re.sub(r"(\s*\n)+", "\n", response.text).replace("\n</span>", "</span>")
    for fragment in expected_fragments:
        # Useful for debugging - uncomment so --pdb gives you 'i'
        # telling you where the first match of the fragment missed.
        # for i in range(len(fragment)):
        #     assert fragment[:i] in html
        assert fragment in html
    for fragment in unexpected_fragments:
        assert fragment not in html


@pytest.mark.asyncio
@pytest.mark.skipif(not shutil.which("rg"), reason="rg executable not found")
async def test_ripgrep_pattern_not_treated_as_flag(datasette):
    response = await datasette.client.get("/-/ripgrep?pattern=-v")
    assert "<title>ripgrep: -v</title>" in response.text
    html = re.sub(r"(\s+\n)+", "\n", response.text)
    assert (
        """
        <h3>{{curlies}}.txt</h3>
        <div style="overflow-x: auto">
        <pre class="match"><a class="line-number" href="/-/ripgrep/view/%7B%7Bcurlies%7D%7D.txt#L1">1   </a><span>File with curlies in the name -v</span></pre>
        </div>
        """.strip()
        in html
    )


@pytest.mark.asyncio
async def test_view_file(datasette):
    response = await datasette.client.get("/-/ripgrep/view/one.txt")
    assert "<h1>one.txt</h1>" in response.text
    assert (
        textwrap.dedent(
            """
    <pre><code id="L1" data-line="1">Hello</code>
    <code id="L2" data-line="2">There</code>
    <code id="L3" data-line="3">This</code>
    <code id="L4" data-line="4">Is a.test file</code>
    </pre>"""
        )
        in response.text
    )


@pytest.mark.asyncio
@pytest.mark.parametrize(
    "path,expected_status,expected_message",
    (
        ("three.txt", 404, "File not found: three.txt"),
        ("..%2F..%2F..%2Ftmp%2Fblah.txt", 403, "File must be inside path directory"),
    ),
)
async def test_view_file_errors(datasette, path, expected_status, expected_message):
    response = await datasette.client.get("/-/ripgrep/view/" + path)
    assert response.status_code == expected_status
    assert expected_message in response.text


@pytest.mark.asyncio
async def test_view_file_with_curlies(datasette):
    response = await datasette.client.get("/-/ripgrep/view/%7B%7Bcurlies%7D%7D.txt")
    assert response.status_code == 200
    assert "File with curlies in the name" in response.text


@pytest.mark.asyncio
async def test_menu_link(datasette):
    response = await datasette.client.get("/")
    assert response.status_code == 200
    assert '<li><a href="/-/ripgrep">ripgrep search</a></li>' in response.text


@pytest.mark.asyncio
@pytest.mark.parametrize(
    "config_overrides,authenticated,path,expected_status",
    [
        # Deny all access
        ({"allow": False}, False, "/-/ripgrep", 403),
        ({"allow": False}, True, "/-/ripgrep", 403),
        ({"allow": False}, False, "/-/ripgrep/view/one.txt", 403),
        ({"allow": False}, True, "/-/ripgrep/view/one.txt", 403),
        # Allow all access
        ({"allow": True}, False, "/-/ripgrep", 200),
        ({"allow": True}, True, "/-/ripgrep", 200),
        ({"allow": True}, False, "/-/ripgrep/view/one.txt", 200),
        ({"allow": True}, True, "/-/ripgrep/view/one.txt", 200),
        # Allow only to logged in user
        ({"allow": {"id": "user"}}, False, "/-/ripgrep", 403),
        ({"allow": {"id": "user"}}, True, "/-/ripgrep", 200),
        ({"allow": {"id": "user"}}, False, "/-/ripgrep/view/one.txt", 403),
        ({"allow": {"id": "user"}}, True, "/-/ripgrep/view/one.txt", 200),
    ],
)
async def test_permissions(src, config_overrides, authenticated, path, expected_status):
    datasette = Datasette(
        [],
        memory=True,
        config={
            **{
                "plugins": {
                    "datasette-ripgrep": {
                        "path": str(src),
                    }
                },
            },
            **config_overrides,
        },
    )
    cookies = {}
    if authenticated:
        cookies["ds_actor"] = datasette.sign({"a": {"id": "user"}}, "actor")
    response = await datasette.client.get(path, cookies=cookies)
    assert response.status_code == expected_status


@pytest.mark.asyncio
@pytest.mark.parametrize("configured", (True, False))
async def test_configuration(configured):
    config = {}
    if configured:
        config = {
            "plugins": {
                "datasette-ripgrep": {
                    "path": "/tmp",
                }
            }
        }
    datasette = Datasette(memory=True, config=config)
    response = await datasette.client.get("/-/ripgrep")
    if not configured:
        assert response.status_code == 500
        assert "The path plugin configuration is required." in response.text
    else:
        assert response.status_code == 200
    # And check the navigation menu
    response2 = await datasette.client.get("/")
    assert response2.status_code == 200
    menu_fragment = '<li><a href="/-/ripgrep">ripgrep search</a></li>'
    if not configured:
        assert menu_fragment not in response2.text
    else:
        assert menu_fragment in response2.text



from setuptools import setup
import os

VERSION = "0.9a0"


def get_long_description():
    with open(
        os.path.join(os.path.dirname(os.path.abspath(__file__)), "README.md"),
        encoding="utf8",
    ) as fp:
        return fp.read()


setup(
    name="datasette-ripgrep",
    description="Web interface for searching your code using ripgrep, built as a Datasette plugin",
    long_description=get_long_description(),
    long_description_content_type="text/markdown",
    author="Simon Willison",
    url="https://github.com/simonw/datasette-ripgrep",
    project_urls={
        "Issues": "https://github.com/simonw/datasette-ripgrep/issues",
        "CI": "https://github.com/simonw/datasette-ripgrep/actions",
        "Changelog": "https://github.com/simonw/datasette-ripgrep/releases",
    },
    license="Apache License, Version 2.0",
    classifiers=[
        "Framework :: Datasette"
    ],
    version=VERSION,
    packages=["datasette_ripgrep"],
    entry_points={"datasette": ["ripgrep = datasette_ripgrep"]},
    package_data={"datasette_ripgrep": ["templates/*.html"]},
    install_requires=["datasette>=1.0a21"],
    extras_require={"test": ["pytest", "pytest-asyncio", "httpx"]},
    tests_require=["datasette-ripgrep[test]"],
    python_requires=">=3.10",
)
